@@ -3,12 +3,10 @@ Built-in block-level token classes.
 """
 
 import re
-import sys
 from itertools import zip_longest
 import mistletoe.block_tokenizer as tokenizer
-from mistletoe import span_token
+from mistletoe import token, span_token
 from mistletoe.core_tokens import (
-        is_link_label,
         follows,
         shift_whitespace,
         whitespace,
@@ -24,15 +22,6 @@ __all__ = ['BlockCode', 'Heading', 'Quote', 'CodeFence', 'ThematicBreak',
            'List', 'Table', 'Footnote', 'Paragraph']
 
 
-"""
-Stores a reference to the current document token.
-
-When parsing, footnote entries will be stored in the document by
-accessing this pointer.
-"""
-_root_node = None
-
-
 def tokenize(lines):
     """
     A wrapper around block_tokenizer.tokenize. Pass in all block-level
@@ -80,7 +69,7 @@ def reset_tokens():
     _token_types = [globals()[cls_name] for cls_name in __all__]
 
 
-class BlockToken(object):
+class BlockToken(token.Token):
     """
     Base class for block-level tokens. Recursively parse inner tokens.
 
@@ -138,34 +127,37 @@ class BlockToken(object):
 class Document(BlockToken):
     """
     Document token.
+    This is a container block token. Its children are block tokens - container or leaf ones.
+
+    Attributes:
+        footnotes (dictionary): link reference definitions.
     """
+
     def __init__(self, lines):
         if isinstance(lines, str):
             lines = lines.splitlines(keepends=True)
         lines = [line if line.endswith('\n') else '{}\n'.format(line) for line in lines]
         self.footnotes = {}
-        global _root_node
-        _root_node = self
-        span_token._root_node = self
+        token._root_node = self
         self.children = tokenize(lines)
-        span_token._root_node = None
-        _root_node = None
+        token._root_node = None
 
 
 class Heading(BlockToken):
     """
-    Heading token. (["### some heading ###\\n"])
-    Boundary between span-level and block-level tokens.
+    ATX heading token. (["### some heading ###\\n"])
+    This is a leaf block token. Its children are inline (span) tokens.
 
     Attributes:
         level (int): heading level.
-        children (list): inner tokens.
     """
-    pattern = re.compile(r' {0,3}(#{1,6})(?:\n|\s+?(.*?)(?:\n|\s+?#+\s*?$))')
+    repr_attributes = ("level",)
+    pattern = re.compile(r' {0,3}(#{1,6})(?:\n|\s+?(.*?)(\n|\s+?#+\s*?$))')
     level = 0
     content = ''
+
     def __init__(self, match):
-        self.level, content = match
+        self.level, content, self.closing_sequence = match
         super().__init__(content, span_token.tokenize_inner)
 
     @classmethod
@@ -177,21 +169,29 @@ class Heading(BlockToken):
         cls.content = (match_obj.group(2) or '').strip()
         if set(cls.content) == {'#'}:
             cls.content = ''
+        cls.closing_sequence = (match_obj.group(3) or '').strip()
         return True
 
     @classmethod
     def read(cls, lines):
         next(lines)
-        return cls.level, cls.content
+        return cls.level, cls.content, cls.closing_sequence
 
 class SetextHeading(BlockToken):
     """
-    Setext headings.
-    
+    Setext heading token.
+    This is a leaf block token. Its children are inline (span) tokens.
+
     Not included in the parsing process, but called by Paragraph.__new__.
+
+    Attributes:
+        level (int): heading level.
     """
+    repr_attributes = ("level",)
+
     def __init__(self, lines):
-        self.level = 1 if lines.pop().lstrip().startswith('=') else 2
+        self.underline = lines.pop().rstrip()
+        self.level = 1 if self.underline.endswith('=') else 2
         content = '\n'.join([line.strip() for line in lines])
         super().__init__(content, span_token.tokenize_inner)
 
@@ -206,7 +206,8 @@ class SetextHeading(BlockToken):
 
 class Quote(BlockToken):
     """
-    Quote token. (["> # heading\\n", "> paragraph\\n"])
+    Block quote token. (["> # heading\\n", "> paragraph\\n"])
+    This is a container block token. Its children are block tokens - container or leaf ones.
     """
     def __init__(self, parse_buffer):
         # span-level tokenizing happens here.
@@ -287,7 +288,7 @@ class Quote(BlockToken):
 class Paragraph(BlockToken):
     """
     Paragraph token. (["some\\n", "continuous\\n", "lines\\n"])
-    Boundary between span-level and block-level tokens.
+    This is a leaf block token. Its children are inline (span) tokens.
     """
     setext_pattern = re.compile(r' {0,3}(=|-)+ *$')
     parse_setext = True  # can be disabled by Quote
@@ -317,14 +318,13 @@ class Paragraph(BlockToken):
                 and not Quote.start(next_line)):
 
             # check if next_line starts List
-            list_pair = ListItem.parse_marker(next_line)
-            if (len(next_line) - len(next_line.lstrip()) < 4
-                    and list_pair is not None):
-                prepend, leader = list_pair
-                # non-empty list item
-                if next_line[:prepend].endswith(' '):
-                    # unordered list, or ordered list starting from 1
-                    if not leader[:-1].isdigit() or leader[:-1] == '1':
+            marker_tuple = ListItem.parse_marker(next_line)
+            if (marker_tuple is not None):
+                _, leader, content = marker_tuple
+                # to break a paragraph, the first line may not be empty,
+                # and the list must be unordered or start from 1.
+                if not content.strip() == '':
+                    if not leader[0].isdigit() or leader in ['1.', '1)']:
                         break
 
             # check if next_line starts HTMLBlock other than type 7
@@ -353,16 +353,22 @@ class Paragraph(BlockToken):
 
 class BlockCode(BlockToken):
     """
-    Indented code.
+    Indented code block token.
+    This is a leaf block token with a single child of type span_token.RawText.
 
     Attributes:
-        children (list): contains a single span_token.RawText token.
         language (str): always the empty string.
     """
+    repr_attributes = ("language",)
     def __init__(self, lines):
         self.language = ''
         self.children = (span_token.RawText(''.join(lines).strip('\n')+'\n'),)
 
+    @property
+    def content(self):
+        """Returns the code block content."""
+        return self.children[0].content
+
     @staticmethod
     def start(line):
         return line.replace('\t', '    ', 1).startswith('    ')
@@ -370,14 +376,20 @@ class BlockCode(BlockToken):
     @classmethod
     def read(cls, lines):
         line_buffer = []
+        trailing_blanks = 0
         for line in lines:
             if line.strip() == '':
                 line_buffer.append(line.lstrip(' ') if len(line) < 5 else line[4:])
+                trailing_blanks = trailing_blanks + 1 if line == '\n' else 0
                 continue
             if not line.replace('\t', '    ', 1).startswith('    '):
                 lines.backstep()
                 break
             line_buffer.append(cls.strip(line))
+            trailing_blanks = 0
+        for _ in range(trailing_blanks):
+            line_buffer.pop()
+            lines.backstep()
         return line_buffer
 
     @staticmethod
@@ -397,29 +409,40 @@ class BlockCode(BlockToken):
 
 class CodeFence(BlockToken):
     """
-    Code fence. (["```sh\\n", "rm -rf /", ..., "```"])
-    Boundary between span-level and block-level tokens.
+    Fenced code block token. (["```sh\\n", "rm -rf /", ..., "```"])
+    This is a leaf block token with a single child of type span_token.RawText.
 
     Attributes:
-        children (list): contains a single span_token.RawText token.
         language (str): language of code block (default to empty).
     """
-    pattern = re.compile(r'( {0,3})((?:`|~){3,}) *(\S*)')
+    repr_attributes = ("language",)
+    pattern = re.compile(r'( {0,3})(`{3,}|~{3,})( *(\S*)[^\n]*)')
     _open_info = None
+
     def __init__(self, match):
         lines, open_info = match
-        self.language = span_token.EscapeSequence.strip(open_info[2])
+        self.indentation = open_info[0]
+        self.delimiter = open_info[1]
+        self.info_string = open_info[2]
+        self.language = span_token.EscapeSequence.strip(open_info[3])
         self.children = (span_token.RawText(''.join(lines)),)
 
+    @property
+    def content(self):
+        """Returns the code block content."""
+        return self.children[0].content
+
     @classmethod
     def start(cls, line):
         match_obj = cls.pattern.match(line)
         if not match_obj:
             return False
-        prepend, leader, lang = match_obj.groups()
-        if leader[0] in lang or leader[0] in line[match_obj.end():]:
+        prepend, leader, info_string, lang = match_obj.groups()
+        # info strings for backtick code blocks may not contain backticks,
+        # but info strings for tilde code blocks may contain both tildes and backticks.
+        if leader[0] == '`' and '`' in info_string:
             return False
-        cls._open_info = len(prepend), leader, lang
+        cls._open_info = len(prepend), leader, info_string, lang
         return True
 
     @classmethod
@@ -442,12 +465,13 @@ class CodeFence(BlockToken):
 class List(BlockToken):
     """
     List token.
+    This is a container block token. Its children are ListItem tokens.
 
     Attributes:
-        children (list): a list of ListItem tokens.
         loose (bool): whether the list is loose.
         start (NoneType or int): None if unordered, starting number if ordered.
     """
+    repr_attributes = ("loose", "start")
     pattern = re.compile(r' {0,3}(?:\d{0,9}[.)]|[+\-*])(?:[ \t]*$|[ \t]+)')
     def __init__(self, matches):
         self.children = [ListItem(*match) for match in matches]
@@ -494,10 +518,20 @@ class List(BlockToken):
 
 class ListItem(BlockToken):
     """
-    List items. Not included in the parsing process, but called by List.
-    """
+    List item token.
+    This is a container block token. Its children are block tokens - container or leaf ones.
 
-    pattern = re.compile(r'\s*(\d{0,9}[.)]|[+\-*])(\s*$|\s+)')
+    Not included in the parsing process, but called by List.
+
+    Attributes:
+        leader (string): a bullet list marker or an ordered list marker.
+        prepend (int): the start position of the content, i.e., the indentation required
+                       for continuation lines.
+        loose (bool): whether the list is loose.
+    """
+    repr_attributes = ("leader", "prepend", "loose")
+    pattern = re.compile(r' {0,3}(\d{0,9}[.)]|[+\-*])($|\s+)')
+    continuation_pattern = re.compile(r'([ \t]*)(\S.*\n|\n)')
 
     def __init__(self, parse_buffer, prepend, leader):
         self.leader = leader
@@ -505,9 +539,23 @@ class ListItem(BlockToken):
         self.children = tokenizer.make_tokens(parse_buffer)
         self.loose = parse_buffer.loose
 
-    @staticmethod
-    def in_continuation(line, prepend):
-        return line.strip() == '' or len(line) - len(line.lstrip()) >= prepend
+    @classmethod
+    def parse_continuation(cls, line, prepend):
+        """
+        Returns content (i.e. the line with the prepend stripped off) iff the line
+        is a valid continuation line for a list item with the given prepend length,
+        otherwise None.
+
+        Note that the list item may still continue even if this test doesn't pass
+        due to lazy continuation.
+        """
+        match_obj = cls.continuation_pattern.match(line)
+        if match_obj is None:
+            return None
+        if match_obj.group(2) == '\n':
+            return '\n'
+        expanded_spaces = match_obj.group(1).expandtabs(4)
+        return expanded_spaces[prepend:] + match_obj.group(2) if len(expanded_spaces) >= prepend else None
 
     @staticmethod
     def other_token(line):
@@ -519,89 +567,93 @@ class ListItem(BlockToken):
     @classmethod
     def parse_marker(cls, line):
         """
-        Returns a pair (prepend, leader) iff the line has a valid leader.
+        Returns a tuple (prepend, leader, content) iff the line has a valid leader and at
+        least one space separating leader and content, or if the content is empty, in which
+        case there need not be any spaces.
+        The return value is None if the line doesn't have a valid marker.
+
+        The leader is a bullet list marker, or an ordered list marker.
+
+        The prepend is the start position of the content, i.e., the indentation required
+        for continuation lines.
         """
         match_obj = cls.pattern.match(line)
         if match_obj is None:
-            return None        # no valid leader
+            return None
+        prepend = len(match_obj.group(0).expandtabs(4))
         leader = match_obj.group(1)
-        content = match_obj.group(0).replace(leader+'\t', leader+'   ', 1)
-        # reassign prepend and leader
-        prepend = len(content)
-        if prepend == len(line.rstrip('\n')):
-            prepend = match_obj.end(1) + 1
-        else:
-            spaces = match_obj.group(2)
-            if spaces.startswith('\t'):
-                spaces = spaces.replace('\t', '   ', 1)
-            spaces = spaces.replace('\t', '    ')
-            n_spaces = len(spaces)
-            if n_spaces > 4:
-                prepend = match_obj.end(1) + 1
-        return prepend, leader
+        content = line[match_obj.end(0):]
+        n_spaces = prepend - match_obj.end(1)
+        if n_spaces > 4:
+            # if there are more than 4 spaces after the leader, we treat them as part of the content
+            # with the exception of the first (marker separator) space.
+            prepend -= n_spaces - 1
+            content = ' ' * (n_spaces - 1) + content
+        return prepend, leader, content
 
     @classmethod
     def read(cls, lines, prev_marker=None):
         next_marker = None
         lines.anchor()
-        prepend = -1
-        leader = None
         line_buffer = []
 
         # first line
         line = next(lines)
-        prepend, leader = prev_marker if prev_marker else cls.parse_marker(line)
-        line = line.replace(leader+'\t', leader+'   ', 1).replace('\t', '    ')
-        empty_first_line = line[prepend:].strip() == ''
-        if not empty_first_line:
-            line_buffer.append(line[prepend:])
         next_line = lines.peek()
-        if empty_first_line and next_line is not None and next_line.strip() == '':
-            parse_buffer = tokenizer.tokenize_block([next(lines)], _token_types)
-            next_line = lines.peek()
-            if next_line is not None:
-                marker_info = cls.parse_marker(next_line)
-                if marker_info is not None:
-                    next_marker = marker_info
-            return (parse_buffer, prepend, leader), next_marker
+        prepend, leader, content = prev_marker if prev_marker else cls.parse_marker(line)
+        if content.strip() == '':
+            # item starting with a blank line: look for the next non-blank line
+            prepend = len(leader) + 1
+            blanks = 1
+            while next_line is not None and next_line.strip() == '':
+                blanks += 1
+                next(lines)
+                next_line = lines.peek()
+            # if the line following the list marker is also empty, then this is an empty
+            # list item.
+            if blanks > 1:
+                parse_buffer = tokenizer.ParseBuffer()
+                parse_buffer.loose = True
+                next_marker = cls.parse_marker(next_line) if next_line is not None else None
+                return (parse_buffer, prepend, leader), next_marker
+        else:
+            line_buffer.append(content)
 
-        # loop
-        newline = 0
+        # loop over the following lines, looking for the end of the list item
+        newline_count = 0
         while True:
-            # no more lines
             if next_line is None:
-                # strip off newlines
-                if newline:
+                # list item ends here because we have reached the end of content
+                if newline_count:
                     lines.backstep()
-                    del line_buffer[-newline:]
+                    del line_buffer[-newline_count:]
                 break
-            next_line = next_line.replace('\t', '    ')
-            # not in continuation
-            if not cls.in_continuation(next_line, prepend):
-                # directly followed by another token
+
+            continuation = cls.parse_continuation(next_line, prepend)
+            if not continuation:
+                # the line doesn't have the indentation to show that it belongs to
+                # the list item, but it should be included anyway by lazy continuation...
+                # ...unless it's the start of another token
                 if cls.other_token(next_line):
-                    if newline:
+                    if newline_count:
                         lines.backstep()
-                        del line_buffer[-newline:]
+                        del line_buffer[-newline_count:]
                     break
-                # next_line is a new list item
+                # ...or it's a new list item
                 marker_info = cls.parse_marker(next_line)
                 if marker_info is not None:
                     next_marker = marker_info
                     break
-                # not another item, has newlines -> not continuation
-                if newline:
+                # ...or the line above it was blank
+                if newline_count:
                     lines.backstep()
-                    del line_buffer[-newline:]
+                    del line_buffer[-newline_count:]
                     break
+                continuation = next_line
+
+            line_buffer.append(continuation)
+            newline_count = newline_count + 1 if continuation == '\n' else 0
             next(lines)
-            line = next_line
-            stripped = line.lstrip(' ')
-            diff = len(line) - len(stripped)
-            if diff > prepend:
-                stripped = ' ' * (diff - prepend) + stripped
-            line_buffer.append(stripped)
-            newline = newline + 1 if next_line.strip() == '' else 0
             next_line = lines.peek()
 
         # block-level tokens are parsed here, so that footnotes can be
@@ -613,12 +665,13 @@ class ListItem(BlockToken):
 class Table(BlockToken):
     """
     Table token.
+    This is a container block token. Its children are TableRow tokens.
 
     Attributes:
-        has_header (bool): whether table has header row.
+        header: header row (TableRow).
         column_align (list): align options for each column (default to [None]).
-        children (list): inner tokens (TableRows).
     """
+    repr_attributes = ("column_align",)
     def __init__(self, lines):
         if '---' in lines[1]:
             self.column_align = [self.parse_align(column)
@@ -626,6 +679,7 @@ class Table(BlockToken):
             self.header = TableRow(lines[0], self.column_align)
             self.children = [TableRow(line, self.column_align) for line in lines[2:]]
         else:
+            # note: not reachable, because read() guarantees the presence of three dashes
             self.column_align = [None]
             self.children = [TableRow(line) for line in lines]
 
@@ -672,28 +726,38 @@ class Table(BlockToken):
 
 class TableRow(BlockToken):
     """
-    Table row token.
+    Table row token. Supports escaped pipes in table cells (for primary use within code spans).
+    This is a container block token. Its children are TableCell tokens.
 
     Should only be called by Table.__init__().
+
+    Attributes:
+        row_align (list): align options for each column (default to [None]).
     """
+    repr_attributes = ("row_align",)
+    # Note: Python regex requires fixed-length look-behind,
+    # so we cannot use a more precise alternative: r"(?<!\\(?:\\\\)*)(\|)"
+    split_pattern = re.compile(r"(?<!\\)\|")
+    escaped_pipe_pattern = re.compile(r"(?<!\\)(\\\\)*\\\|")
+
     def __init__(self, line, row_align=None):
         self.row_align = row_align or [None]
-        cells = filter(None, line.strip().split('|'))
-        self.children = [TableCell(cell.strip() if cell else '', align)
+        cells = filter(None, self.split_pattern.split(line.strip()))
+        self.children = [TableCell(self.escaped_pipe_pattern.sub('\\1|', cell.strip()) if cell else '', align)
                          for cell, align in zip_longest(cells, self.row_align)]
 
 
 class TableCell(BlockToken):
     """
     Table cell token.
-    Boundary between span-level and block-level tokens.
+    This is a leaf block token. Its children are inline (span) tokens.
 
     Should only be called by TableRow.__init__().
 
     Attributes:
         align (bool): align option for current cell (default to None).
-        children (list): inner (span-)tokens.
     """
+    repr_attributes = ("align",)
     def __init__(self, content, align=None):
         self.align = align
         super().__init__(content, span_token.tokenize_inner)
@@ -701,13 +765,13 @@ class TableCell(BlockToken):
 
 class Footnote(BlockToken):
     """
-    Footnote token.
+    Footnote token. One or more "link reference definitions" according to the CommonMark spec.
+    This is a leaf block token without children.
 
-    The constructor returns None, because the footnote information
-    is stored in Footnote.read.
+    The constructor returns None, because available footnote definitions are parsed
+    and stored into the root node within `Footnote.read()`. We don't put instances of
+    this class into the resulting AST.
     """
-    label_pattern = re.compile(r'[ \n]{0,3}\[(.+?)\]', re.DOTALL)
-
     def __new__(cls, _):
         return None
 
@@ -726,74 +790,116 @@ class Footnote(BlockToken):
         offset = 0
         matches = []
         while offset < len(string) - 1:
-            match_info = cls.match_reference(lines, string, offset)
+            match_info = cls.match_reference(string, offset)
             if match_info is None:
+                # backtrack the lines that have not been consumed
+                lines._index -= string[offset:].count('\n')
                 break
             offset, match = match_info
             matches.append(match)
-        cls.append_footnotes(matches, _root_node)
+        cls.append_footnotes(matches, token._root_node)
         return matches or None
 
     @classmethod
-    def match_reference(cls, lines, string, offset):
+    def match_reference(cls, string, offset):
+        # up to three spaces, "[", label, "]"
         match_info = cls.match_link_label(string, offset)
         if not match_info:
-            cls.backtrack(lines, string, offset)
             return None
         _, label_end, label = match_info
 
+        # ":"
         if not follows(string, label_end-1, ':'):
-            cls.backtrack(lines, string, offset)
             return None
 
-        match_info = cls.match_link_dest(string, label_end)
+        # optional spaces or tabs (including up to one line ending)
+        dest_start = shift_whitespace(string, label_end + 1)
+        if dest_start == len(string):
+            return None
+
+        # link destination
+        match_info = cls.match_link_dest(string, dest_start)
         if not match_info:
-            cls.backtrack(lines, string, offset)
             return None
         _, dest_end, dest = match_info
+        dest_type = "angle_uri" if string[dest_start] == "<" else "uri"
 
-        match_info = cls.match_link_title(string, dest_end)
-        if not match_info:
-            cls.backtrack(lines, string, dest_end)
+        # either of:
+        # 1) optional spaces or tabs and then a line break to finish the link reference definition.
+        # 2) optional spaces or tabs (including up to one line ending) followed by a title.
+        # in any case, if the destination is followed directly by non-whitespace, then it's not
+        # a valid link reference definition.
+        title_start = shift_whitespace(string, dest_end)
+        if title_start == dest_end and title_start < len(string):
             return None
+
+        # link title
+        match_info = cls.match_link_title(string, title_start)
+        if not match_info:
+            # no valid title found. if there was a line break following the destination,
+            # we still have a valid link reference definition. otherwise not.
+            eol_pos = string[dest_end:title_start].find("\n")
+            if eol_pos >= 0:
+                return dest_end + eol_pos + 1, (label, dest, "", dest_type, None)
+            else:
+                return None
         _, title_end, title = match_info
 
-        return title_end, (label, dest, title)
+        # optional spaces or tabs. final line ending.
+        line_end = title_end
+        while line_end < len(string):
+            if string[line_end] == '\n':
+                title_delimiter = string[title_start] if title_start < title_end else None
+                return line_end + 1, (label, dest, title, dest_type, title_delimiter)
+            elif string[line_end] in whitespace:
+                line_end += 1
+            else:
+                break
+
+        # non-whitespace found on the same line as the title, making it invalid.
+        # if there was a line break following the destination,
+        # we still have a valid link reference definition. otherwise not.
+        eol_pos = string[dest_end:title_start].find("\n")
+        if eol_pos >= 0:
+            return dest_end + eol_pos + 1, (label, dest, "", dest_type, None)
+        else:
+            return None
 
     @classmethod
     def match_link_label(cls, string, offset):
+        """
+        Matches: up to three spaces, "[", label, "]".
+        """
         start = -1
-        end = -1
         escaped = False
         for i, c in enumerate(string[offset:], start=offset):
-            if c == '\\' and not escaped:
+            if escaped:
+                escaped = False
+            elif c == '\\':
                 escaped = True
-            elif c == '[' and not escaped:
+            elif c == '[':
                 if start == -1:
                     start = i
                 else:
                     return None
-            elif c == ']' and not escaped:
-                end = i
-                label = string[start+1:end]
+            elif c == ']':
+                label = string[start+1:i]
                 if label.strip() != '':
-                    return start, end+1, label
+                    return start, i+1, label
+                return None
+            # only spaces allowed before the opening bracket
+            if start == -1 and not (c == " " and i - offset < 3):
                 return None
-            elif escaped:
-                escaped = False
         return None
 
     @classmethod
     def match_link_dest(cls, string, offset):
-        offset = shift_whitespace(string, offset+1)
-        if offset == len(string):
-            return None
         if string[offset] == '<':
             escaped = False
             for i, c in enumerate(string[offset+1:], start=offset+1):
                 if c == '\\' and not escaped:
                     escaped = True
-                elif c == ' ' or c == '\n' or (c == '<' and not escaped):
+                elif c == '\n' or (c == '<' and not escaped):
                     return None
                 elif c == '>' and not escaped:
                     return offset, i+1, string[offset+1:i]
@@ -823,56 +929,45 @@ class Footnote(BlockToken):
 
     @classmethod
     def match_link_title(cls, string, offset):
-        new_offset = shift_whitespace(string, offset)
-        if (new_offset == len(string)
-                or '\n' in string[offset:new_offset]
-                and string[new_offset] == '['):
-            return offset, new_offset, ''
-        if string[new_offset] == '"':
+        if offset == len(string):
+            return None
+        if string[offset] == '"':
             closing = '"'
-        elif string[new_offset] == "'":
+        elif string[offset] == "'":
             closing = "'"
-        elif string[new_offset] == '(':
+        elif string[offset] == '(':
             closing = ')'
-        elif '\n' in string[offset:new_offset]:
-            return offset, offset, ''
         else:
             return None
-        offset = new_offset
         escaped = False
         for i, c in enumerate(string[offset+1:], start=offset+1):
             if c == '\\' and not escaped:
                 escaped = True
             elif c == closing and not escaped:
-                new_offset = shift_whitespace(string, i+1)
-                if '\n' not in string[i+1:new_offset]:
-                    return None
-                return offset, new_offset, string[offset+1:i]
+                return offset, i+1, string[offset+1:i]
             elif escaped:
                 escaped = False
         return None
 
     @staticmethod
     def append_footnotes(matches, root):
-        for key, dest, title in matches:
+        for key, dest, title, *_ in matches:
             key = normalize_label(key)
             dest = span_token.EscapeSequence.strip(dest.strip())
             title = span_token.EscapeSequence.strip(title)
             if key not in root.footnotes:
                 root.footnotes[key] = dest, title
 
-    @staticmethod
-    def backtrack(lines, string, offset):
-        lines._index -= string[offset+1:].count('\n')
-
 
 class ThematicBreak(BlockToken):
     """
     Thematic break token (a.k.a. horizontal rule.)
+    This is a leaf block token without children.
     """
     pattern = re.compile(r' {0,3}(?:([-_*])\s*?)(?:\1\s*?){2,}$')
-    def __init__(self, _):
-        pass
+
+    def __init__(self, lines):
+        self.line = lines[0].strip('\n')
 
     @classmethod
     def start(cls, line):
@@ -885,26 +980,30 @@ class ThematicBreak(BlockToken):
 
 class HTMLBlock(BlockToken):
     """
-    Block-level HTML tokens.
-
-    Attributes:
-        content (str): literal strings rendered as-is.
+    Block-level HTML token.
+    This is a leaf block token with a single child of type span_token.RawText,
+    which holds the raw HTML content.
     """
     _end_cond = None
-    multiblock = re.compile(r'<(script|pre|style)[ >\n]')
+    multiblock = re.compile(r'<(pre|script|style|textarea)[ >\n]')
     predefined = re.compile(r'<\/?(.+?)(?:\/?>|[ \n])')
     custom_tag = re.compile(r'(?:' + '|'.join((span_token._open_tag,
                                 span_token._closing_tag)) + r')\s*$')
 
     def __init__(self, lines):
-        self.content = ''.join(lines).rstrip('\n')
+        self.children = (span_token.RawText(''.join(lines).rstrip('\n')),)
+
+    @property
+    def content(self):
+        """Returns the raw HTML content."""
+        return self.children[0].content
 
     @classmethod
     def start(cls, line):
         stripped = line.lstrip()
         if len(line) - len(stripped) >= 4:
             return False
-        # rule 1: <pre>, <script> or <style> tags, allow newlines in block
+        # rule 1: HTML tags designed to contain literal content, allow newlines in block
         match_obj = cls.multiblock.match(stripped)
         if match_obj is not None:
             cls._end_cond = '</{}>'.format(match_obj.group(1).casefold())
@@ -948,6 +1047,7 @@ class HTMLBlock(BlockToken):
                     break
             elif line.strip() == '':
                 line_buffer.pop()
+                lines.backstep()
                 break
         return line_buffer
 
