@@ -2,17 +2,14 @@
 HTML renderer for mistletoe.
 """
 
-import re
-import sys
+import html
 from itertools import chain
 from urllib.parse import quote
+from mistletoe import block_token
+from mistletoe import span_token
 from mistletoe.block_token import HTMLBlock
 from mistletoe.span_token import HTMLSpan
 from mistletoe.base_renderer import BaseRenderer
-if sys.version_info < (3, 4):
-    from mistletoe import _html as html
-else:
-    import html
 
 
 class HTMLRenderer(BaseRenderer):
@@ -21,67 +18,67 @@ class HTMLRenderer(BaseRenderer):
 
     See mistletoe.base_renderer module for more info.
     """
-    def __init__(self, *extras):
+    def __init__(self, *extras, html_escape_double_quotes=True, html_escape_single_quotes=False, **kwargs):
         """
         Args:
             extras (list): allows subclasses to add even more custom tokens.
+            html_escape_double_quotes (bool): whether to also escape double
+                quotes when HTML-escaping rendered text.
+            html_escape_single_quotes (bool): whether to also escape single
+                quotes when HTML-escaping rendered text.
+            **kwargs: additional parameters to be passed to the ancestor's
+                constructor.
         """
         self._suppress_ptag_stack = [False]
-        super().__init__(*chain((HTMLBlock, HTMLSpan), extras))
-        # html.entities.html5 includes entitydefs not ending with ';',
-        # CommonMark seems to hate them, so...
-        self._stdlib_charref = html._charref
-        _charref = re.compile(r'&(#[0-9]+;'
-                              r'|#[xX][0-9a-fA-F]+;'
-                              r'|[^\t\n\f <&#;]{1,32};)')
-        html._charref = _charref
+        super().__init__(*chain((HTMLBlock, HTMLSpan), extras), **kwargs)
+        self.html_escape_double_quotes = html_escape_double_quotes
+        self.html_escape_single_quotes = html_escape_single_quotes
 
     def __exit__(self, *args):
         super().__exit__(*args)
-        html._charref = self._stdlib_charref
 
-    def render_to_plain(self, token):
+    def render_to_plain(self, token) -> str:
         if hasattr(token, 'children'):
             inner = [self.render_to_plain(child) for child in token.children]
             return ''.join(inner)
-        return self.escape_html(token.content)
+        return html.escape(token.content)
 
-    def render_strong(self, token):
+    def render_strong(self, token: span_token.Strong) -> str:
         template = '<strong>{}</strong>'
         return template.format(self.render_inner(token))
 
-    def render_emphasis(self, token):
+    def render_emphasis(self, token: span_token.Emphasis) -> str:
         template = '<em>{}</em>'
         return template.format(self.render_inner(token))
 
-    def render_inline_code(self, token):
+    def render_inline_code(self, token: span_token.InlineCode) -> str:
         template = '<code>{}</code>'
-        inner = html.escape(token.children[0].content)
+        inner = self.escape_html_text(token.children[0].content)
         return template.format(inner)
 
-    def render_strikethrough(self, token):
+    def render_strikethrough(self, token: span_token.Strikethrough) -> str:
         template = '<del>{}</del>'
         return template.format(self.render_inner(token))
 
-    def render_image(self, token):
+    def render_image(self, token: span_token.Image) -> str:
         template = '<img src="{}" alt="{}"{} />'
         if token.title:
-            title = ' title="{}"'.format(self.escape_html(token.title))
+            title = ' title="{}"'.format(html.escape(token.title))
         else:
             title = ''
         return template.format(token.src, self.render_to_plain(token), title)
 
-    def render_link(self, token):
+    def render_link(self, token: span_token.Link) -> str:
         template = '<a href="{target}"{title}>{inner}</a>'
         target = self.escape_url(token.target)
         if token.title:
-            title = ' title="{}"'.format(self.escape_html(token.title))
+            title = ' title="{}"'.format(html.escape(token.title))
         else:
             title = ''
         inner = self.render_inner(token)
         return template.format(target=target, title=title, inner=inner)
 
-    def render_auto_link(self, token):
+    def render_auto_link(self, token: span_token.AutoLink) -> str:
         template = '<a href="{target}">{inner}</a>'
         if token.mailto:
             target = 'mailto:{}'.format(token.target)
@@ -90,22 +87,22 @@ class HTMLRenderer(BaseRenderer):
         inner = self.render_inner(token)
         return template.format(target=target, inner=inner)
 
-    def render_escape_sequence(self, token):
+    def render_escape_sequence(self, token: span_token.EscapeSequence) -> str:
         return self.render_inner(token)
 
-    def render_raw_text(self, token):
-        return self.escape_html(token.content)
+    def render_raw_text(self, token: span_token.RawText) -> str:
+        return self.escape_html_text(token.content)
 
     @staticmethod
-    def render_html_span(token):
+    def render_html_span(token: span_token.HTMLSpan) -> str:
         return token.content
 
-    def render_heading(self, token):
+    def render_heading(self, token: block_token.Heading) -> str:
         template = '<h{level}>{inner}</h{level}>'
         inner = self.render_inner(token)
         return template.format(level=token.level, inner=inner)
 
-    def render_quote(self, token):
+    def render_quote(self, token: block_token.Quote) -> str:
         elements = ['<blockquote>']
         self._suppress_ptag_stack.append(False)
         elements.extend([self.render(child) for child in token.children])
@@ -113,21 +110,21 @@ class HTMLRenderer(BaseRenderer):
         elements.append('</blockquote>')
         return '\n'.join(elements)
 
-    def render_paragraph(self, token):
+    def render_paragraph(self, token: block_token.Paragraph) -> str:
         if self._suppress_ptag_stack[-1]:
             return '{}'.format(self.render_inner(token))
         return '<p>{}</p>'.format(self.render_inner(token))
 
-    def render_block_code(self, token):
+    def render_block_code(self, token: block_token.BlockCode) -> str:
         template = '<pre><code{attr}>{inner}</code></pre>'
         if token.language:
-            attr = ' class="{}"'.format('language-{}'.format(self.escape_html(token.language)))
+            attr = ' class="{}"'.format('language-{}'.format(html.escape(token.language)))
         else:
             attr = ''
-        inner = html.escape(token.children[0].content)
+        inner = self.escape_html_text(token.content)
         return template.format(attr=attr, inner=inner)
 
-    def render_list(self, token):
+    def render_list(self, token: block_token.List) -> str:
         template = '<{tag}{attr}>\n{inner}\n</{tag}>'
         if token.start is not None:
             tag = 'ol'
@@ -140,7 +137,7 @@ class HTMLRenderer(BaseRenderer):
         self._suppress_ptag_stack.pop()
         return template.format(tag=tag, attr=attr, inner=inner)
 
-    def render_list_item(self, token):
+    def render_list_item(self, token: block_token.ListItem) -> str:
         if len(token.children) == 0:
             return '<li></li>'
         inner = '\n'.join([self.render(child) for child in token.children])
@@ -152,7 +149,7 @@ class HTMLRenderer(BaseRenderer):
                 inner_template = inner_template[:-1]
         return '<li>{}</li>'.format(inner_template.format(inner))
 
-    def render_table(self, token):
+    def render_table(self, token: block_token.Table) -> str:
         # This is actually gross and I wonder if there's a better way to do it.
         #
         # The primary difficulty seems to be passing down alignment options to
@@ -168,13 +165,13 @@ class HTMLRenderer(BaseRenderer):
         body_rendered = body_template.format(inner=body_inner)
         return template.format(inner=head_rendered+body_rendered)
 
-    def render_table_row(self, token, is_header=False):
+    def render_table_row(self, token: block_token.TableRow, is_header=False) -> str:
         template = '<tr>\n{inner}</tr>\n'
         inner = ''.join([self.render_table_cell(child, is_header)
                          for child in token.children])
         return template.format(inner=inner)
 
-    def render_table_cell(self, token, in_header=False):
+    def render_table_cell(self, token: block_token.TableCell, in_header=False) -> str:
         template = '<{tag}{attr}>{inner}</{tag}>\n'
         tag = 'th' if in_header else 'td'
         if token.align is None:
@@ -188,29 +185,42 @@ class HTMLRenderer(BaseRenderer):
         return template.format(tag=tag, attr=attr, inner=inner)
 
     @staticmethod
-    def render_thematic_break(token):
+    def render_thematic_break(token: block_token.ThematicBreak) -> str:
         return '<hr />'
 
     @staticmethod
-    def render_line_break(token):
+    def render_line_break(token: span_token.LineBreak) -> str:
         return '\n' if token.soft else '<br />\n'
 
     @staticmethod
-    def render_html_block(token):
+    def render_html_block(token: block_token.HTMLBlock) -> str:
         return token.content
 
-    def render_document(self, token):
+    def render_document(self, token: block_token.Document) -> str:
         self.footnotes.update(token.footnotes)
         inner = '\n'.join([self.render(child) for child in token.children])
         return '{}\n'.format(inner) if inner else ''
 
-    @staticmethod
-    def escape_html(raw):
-        return html.escape(html.unescape(raw)).replace('&#x27;', "'")
+    def escape_html_text(self, s: str) -> str:
+        """
+        Like `html.escape()`, but this  looks into the current rendering options
+        to decide which of the quotes (double, single, or both) to escape.
+
+        Intended for escaping text content. To escape content of an attribute,
+        simply call `html.escape()`.
+        """
+        s = s.replace("&", "&amp;") # Must be done first!
+        s = s.replace("<", "&lt;")
+        s = s.replace(">", "&gt;")
+        if self.html_escape_double_quotes:
+            s = s.replace('"', "&quot;")
+        if self.html_escape_single_quotes:
+            s = s.replace('\'', "&#x27;")
+        return s
 
     @staticmethod
-    def escape_url(raw):
+    def escape_url(raw: str) -> str:
         """
         Escape urls to prevent code injection craziness. (Hopefully.)
         """
-        return html.escape(quote(html.unescape(raw), safe='/#:()*?=%@+,&;'))
+        return html.escape(quote(raw, safe='/#:()*?=%@+,&;'))
