@@ -2,9 +2,10 @@
 Built-in span-level token classes.
 """
 
+import html
 import re
 import mistletoe.span_tokenizer as tokenizer
-from mistletoe import core_tokens
+from mistletoe import core_tokens, token
 
 
 """
@@ -14,9 +15,6 @@ __all__ = ['EscapeSequence', 'Strikethrough', 'AutoLink', 'CoreTokens',
            'InlineCode', 'LineBreak', 'RawText']
 
 
-_root_node = None
-
-
 def tokenize_inner(content):
     """
     A wrapper around span_tokenizer.tokenize. Pass in all span-level token
@@ -63,7 +61,7 @@ def reset_tokens():
     _token_types = [globals()[cls_name] for cls_name in __all__]
 
 
-class SpanToken:
+class SpanToken(token.Token):
     parse_inner = True
     parse_group = 1
     precedence = 5
@@ -83,30 +81,43 @@ class SpanToken:
 
 
 class CoreTokens(SpanToken):
+    """
+    Represents core tokens (Strong, Emphasis, Image, Link) during the early stage of parsing.
+    Replaced with objects of the proper classes in the final stage of parsing.
+    """
     precedence = 3
     def __new__(self, match):
         return globals()[match.type](match)
 
     @classmethod
     def find(cls, string):
-        return core_tokens.find_core_tokens(string, _root_node)
+        return core_tokens.find_core_tokens(string, token._root_node)
 
 
 class Strong(SpanToken):
     """
-    Strong tokens. ("**some text**")
+    Strong token. ("**some text**")
+    This is an inline token. Its children are inline (span) tokens.
+    One of the core tokens.
     """
+    def __init__(self, match):
+        self.delimiter = match.delimiter
 
 
 class Emphasis(SpanToken):
     """
-    Emphasis tokens. ("*some text*")
+    Emphasis token. ("*some text*")
+    This is an inline token. Its children are inline (span) tokens.
+    One of the core tokens.
     """
+    def __init__(self, match):
+        self.delimiter = match.delimiter
 
 
 class InlineCode(SpanToken):
     """
-    Inline code tokens. ("`some code`")
+    Inline code token. ("`some code`")
+    This is an inline token with a single child of type RawText.
     """
     pattern = re.compile(r"(?<!\\|`)(?:\\\\)*(`+)(?!`)(.+?)(?<!`)\1(?!`)", re.DOTALL)
     parse_inner = False
@@ -114,7 +125,12 @@ class InlineCode(SpanToken):
 
     def __init__(self, match):
         content = match.group(self.parse_group)
-        self.children = (RawText(' '.join(re.split('[ \n]+', content.strip()))),)
+        self.delimiter = match.group(1)
+        content = content.replace('\n', ' ')
+        self.padding = " " if not content.isspace() and content.startswith(" ") and content.endswith(" ") else ""
+        if self.padding:
+            content = content[1:-1]
+        self.children = (RawText(content),)
 
     @classmethod
     def find(cls, string):
@@ -125,44 +141,63 @@ class InlineCode(SpanToken):
 
 class Strikethrough(SpanToken):
     """
-    Strikethrough tokens. ("~~some text~~")
+    Strikethrough token. ("~~some text~~")
+    This is an inline token. Its children are inline (span) tokens.
     """
     pattern = re.compile(r"(?<!\\)(?:\\\\)*~~(.+?)~~", re.DOTALL)
 
 
 class Image(SpanToken):
     """
-    Image tokens. ("![alt](src "title")")
+    Image token. ("![alt](src "title")")
+    This is an inline token. Its children are inline (span) tokens holding the image description.
+    One of the core tokens.
 
     Attributes:
         src (str): image source.
         title (str): image title (default to empty).
+        label (str): link label, for reference links.
     """
+    repr_attributes = ("src", "title")
     def __init__(self, match):
-        self.src = match.group(2).strip()
-        self.title = match.group(3)
+        self.src = EscapeSequence.strip(match.group(2).strip())
+        self.title = EscapeSequence.strip(match.group(3))
+        self.dest_type = getattr(match, "dest_type", None)
+        self.label = getattr(match, "label", None)
+        self.title_delimiter = getattr(match, "title_delimiter", None)
 
 
 class Link(SpanToken):
     """
-    Link tokens. ("[name](target)")
+    Link token. ("[name](target "title")")
+    This is an inline token. Its children are inline (span) tokens holding the link text.
+    One of the core tokens.
 
     Attributes:
         target (str): link target.
+        title (str): link title (default to empty).
+        label (str): link label, for reference links.
     """
+    repr_attributes = ("target", "title")
     def __init__(self, match):
         self.target = EscapeSequence.strip(match.group(2).strip())
         self.title = EscapeSequence.strip(match.group(3))
+        self.dest_type = getattr(match, "dest_type", None)
+        self.label = getattr(match, "label", None)
+        self.title_delimiter = getattr(match, "title_delimiter", None)
 
 
 class AutoLink(SpanToken):
     """
-    Autolink tokens. ("<http://www.google.com>")
+    Autolink token. ("<http://www.google.com>")
+    This is an inline token with a single child of type RawText.
 
     Attributes:
-        children (iterator): a single RawText node for alternative text.
+        children (list): a single RawText node for the link target.
         target (str): link target.
+        mailto (bool): true iff the target looks like an email address, but does not have the "mailto:" prefix.
     """
+    repr_attributes = ("target", "mailto")
     pattern = re.compile(r"(?<!\\)(?:\\\\)*<([A-Za-z][A-Za-z0-9+.-]{1,31}:[^ <>]*?|[A-Za-z0-9.!#$%&'*+/=?^_`{|}~-]+@[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?(?:\.[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*)>")
     parse_inner = False
 
@@ -175,10 +210,11 @@ class AutoLink(SpanToken):
 
 class EscapeSequence(SpanToken):
     """
-    Escape sequences. ("\*")
+    Escape sequence token. ("\\\\*")
+    This is an inline token with a single child of type RawText.
 
     Attributes:
-        children (iterator): a single RawText node for alternative text.
+        children (iterator): a single RawText node containing the escaped character.
     """
     pattern = re.compile(r"\\([!\"#$%&'()*+,-./:;<=>?@\[\\\]^_`{|}~])")
     parse_inner = False
@@ -189,26 +225,31 @@ class EscapeSequence(SpanToken):
 
     @classmethod
     def strip(cls, string):
-        return cls.pattern.sub(r'\1', string)
+        return html.unescape(cls.pattern.sub(r'\1', string))
 
 
 class LineBreak(SpanToken):
     """
-    Hard or soft line breaks.
+    Line break token: hard or soft.
+    This is an inline token without children.
+
+    Attributes:
+        soft (bool): true if this is a soft line break.
     """
+    repr_attributes = ("soft",)
     pattern = re.compile(r'( *|\\)\n')
     parse_inner = False
     parse_group = 0
 
     def __init__(self, match):
-        content = match.group(1)
-        self.soft = not content.startswith(('  ', '\\'))
-        self.content = ''
+        self.content = match.group(1)
+        self.soft = not self.content.startswith(('  ', '\\'))
 
 
 class RawText(SpanToken):
     """
-    Raw text. A leaf node.
+    Raw text token.
+    This is an inline token without children.
 
     RawText is the only token that accepts a string for its constructor,
     instead of a match object. Also, all recursions should bottom out here.
@@ -228,7 +269,7 @@ _tags = {'address', 'article', 'aside', 'base', 'basefont', 'blockquote',
         'ul'}
 
 _tag   = r'[A-Za-z][A-Za-z0-9-]*'
-_attrs = r'(?:\s+[A-Za-z_:][A-Za-z0-9_.:-]*(?:\s*=\s*(?:[^ "\'=<>`]+|\'[^\']*?\'|"[^\"]*?"))?)*'
+_attrs = r'(?:\s+[A-Za-z_:][A-Za-z0-9_.:-]*(?:\s*=\s*(?:[^\s"\'=<>`]+|\'[^\']*?\'|"[^\"]*?"))?)*'
 
 _open_tag    = r'(?<!\\)<' + _tag + _attrs + r'\s*/?>'
 _closing_tag = r'(?<!\\)</' + _tag + r'\s*>'
@@ -240,10 +281,11 @@ _cdata       = r'(?<!\\)<!\[CDATA.+?\]\]>'
 
 class HTMLSpan(SpanToken):
     """
-    Span-level HTML tokens.
+    Span-level HTML token.
+    This is an inline token without children.
 
     Attributes:
-        content (str): literal strings rendered as-is.
+        content (str): the raw HTML content.
     """
     pattern = re.compile('|'.join([_open_tag, _closing_tag, _comment,
                                    _instruction, _declaration, _cdata]),
@@ -251,6 +293,7 @@ class HTMLSpan(SpanToken):
     parse_inner = False
     parse_group = 0
 
+
 # Note: The following XWiki tokens are based on the XWiki Syntax 2.0 (or above; 1.0 was deprecated years ago already).
 
 class XWikiBlockMacroStart(SpanToken):
@@ -275,4 +318,3 @@ class XWikiBlockMacroEnd(SpanToken):
 
 _token_types = []
 reset_tokens()
-
