@@ -1,7 +1,6 @@
 import unittest
 from unittest.mock import patch, call
 from mistletoe import block_token, span_token
-from mistletoe.block_tokenizer import FileWrapper
 
 
 class TestToken(unittest.TestCase):
@@ -60,7 +59,7 @@ class TestSetextHeading(TestToken):
         self.assertIsInstance(next(tokens), block_token.Paragraph)
         self.mock.assert_has_calls([call('some'), call('heading'), call('foobar')])
         with self.assertRaises(StopIteration) as e:
-            token = next(tokens)
+            next(tokens)
 
 
 class TestQuote(unittest.TestCase):
@@ -114,6 +113,11 @@ class TestCodeFence(TestToken):
         arg = 'hey'
         self._test_match(block_token.CodeFence, lines, arg, language='')
 
+    def test_code_fence_with_backticks_and_tildes_in_the_info_string(self):
+        lines = ['~~~ aa ``` ~~~\n', 'foo\n', '~~~\n']
+        arg = 'foo\n'
+        self._test_match(block_token.CodeFence, lines, arg, language='aa')
+
 
 class TestBlockCode(TestToken):
     def test_parse_indented_code(self):
@@ -121,6 +125,11 @@ class TestBlockCode(TestToken):
         arg = 'rm dir\nmkdir test\n'
         self._test_match(block_token.BlockCode, lines, arg, language='')
 
+    def test_parse_indented_code_with_blank_lines(self):
+        lines = ['    chunk1\n', '\n', '    chunk2\n', '  \n', ' \n', ' \n', '    chunk3\n']
+        arg = 'chunk1\n\nchunk2\n\n\n\nchunk3\n'
+        self._test_match(block_token.BlockCode, lines, arg, language='')
+
 
 class TestParagraph(TestToken):
     def test_parse(self):
@@ -142,17 +151,19 @@ class TestParagraph(TestToken):
 class TestListItem(unittest.TestCase):
     def test_parse_marker(self):
         lines = ['- foo\n',
-                 '*    bar\n',
+                 '   *    bar\n',
                  ' + baz\n',
                  '1. item 1\n',
                  '2) item 2\n',
-                 '123456789. item x\n']
+                 '123456789. item x\n',
+                 '*\n']
         for line in lines:
             self.assertTrue(block_token.ListItem.parse_marker(line))
         bad_lines = ['> foo\n',
                      '1item 1\n',
                      '2| item 2\n',
-                     '1234567890. item x\n']
+                     '1234567890. item x\n',
+                     '    * too many spaces\n']
         for line in bad_lines:
             self.assertFalse(block_token.ListItem.parse_marker(line))
 
@@ -177,7 +188,6 @@ class TestListItem(unittest.TestCase):
         lines = ['- foo\n',
                  '  - bar\n',
                  '    - baz\n']
-        f = FileWrapper(lines)
         ptoken, ltoken = block_token.tokenize(lines)[0].children[0].children
         self.assertIsInstance(ptoken, block_token.Paragraph)
         self.assertIsInstance(ltoken, block_token.List)
@@ -195,7 +205,6 @@ class TestListItem(unittest.TestCase):
                  '  \n',
                  '  baz\n'
                  '  ~~~\n']
-        f = FileWrapper(lines)
         list_item = block_token.tokenize(lines)[0].children[0]
         self.assertEqual(list_item.loose, False)
 
@@ -203,10 +212,84 @@ class TestListItem(unittest.TestCase):
         lines = ['- foo\n',
                  '\n',
                  '# bar\n']
-        f = FileWrapper(lines)
         list_item = block_token.tokenize(lines)[0].children[0]
         self.assertEqual(list_item.loose, False)
 
+    def test_tabbed_list_items(self):
+        # according to the CommonMark spec:
+        # in contexts where spaces help to define block structure, tabs behave as if they
+        # were replaced by spaces with a tab stop of 4 characters.
+        lines = ['title\n',
+                 '*\ttabbed item long line\n',
+                 '\n', # break lazy continuation
+                 '    continuation 1\n',
+                 '*   second list item\n',
+                 '\n', # break lazy continuation
+                 '\tcontinuation 2\n']
+        tokens = block_token.tokenize(lines)
+        self.assertEqual(len(tokens), 2)
+        self.assertIsInstance(tokens[0], block_token.Paragraph)
+        self.assertIsInstance(tokens[1], block_token.List)
+        self.assertTrue('tabbed item long line' in tokens[1].children[0])
+        self.assertTrue('continuation 1' in tokens[1].children[0])
+        self.assertTrue('second list item' in tokens[1].children[1])
+        self.assertTrue('continuation 2' in tokens[1].children[1])
+
+    def test_list_items_starting_with_blank_line(self):
+        lines = ['-\n',
+                 '  foo\n',
+                 '-\n',
+                 '  ```\n',
+                 '  bar\n',
+                 '  ```\n',
+                 '-\n',
+                 '      baz\n']
+        tokens = block_token.tokenize(lines)
+        self.assertEqual(len(tokens), 1)
+        self.assertIsInstance(tokens[0], block_token.List)
+        self.assertIsInstance(tokens[0].children[0].children[0], block_token.Paragraph)
+        self.assertIsInstance(tokens[0].children[1].children[0], block_token.CodeFence)
+        self.assertIsInstance(tokens[0].children[2].children[0], block_token.BlockCode)
+        self.assertTrue('foo' in tokens[0].children[0].children[0])
+        self.assertEqual('bar\n', tokens[0].children[1].children[0].children[0].content)
+        self.assertEqual('baz\n', tokens[0].children[2].children[0].children[0].content)
+
+    def test_a_list_item_may_begin_with_at_most_one_blank_line(self):
+        lines = ['-\n',
+                 '\n',
+                 '  foo\n']
+        tokens = block_token.tokenize(lines)
+        self.assertEqual(len(tokens), 2)
+        self.assertIsInstance(tokens[0], block_token.List)
+        self.assertIsInstance(tokens[1], block_token.Paragraph)
+        self.assertTrue('foo' in tokens[1].children[0])
+
+    def test_empty_list_item_in_the_middle(self):
+        lines = ['* a\n',
+                 '*\n',
+                 '\n',
+                 '* c\n']
+        tokens = block_token.tokenize(lines)
+        self.assertEqual(len(tokens), 1)
+        self.assertIsInstance(tokens[0], block_token.List)
+        self.assertEqual(len(tokens[0].children), 3)
+        self.assertTrue(tokens[0].loose)
+
+    def test_list_with_code_block(self):
+        lines = ['1.      indented code\n',
+                 '\n',
+                 '   paragraph\n',
+                 '\n',
+                 '       more code\n']
+        tokens = block_token.tokenize(lines)
+        self.assertEqual(len(tokens), 1)
+        self.assertIsInstance(tokens[0], block_token.List)
+        self.assertEqual(len(tokens[0].children), 1)
+        self.assertIsInstance(tokens[0].children[0].children[0], block_token.BlockCode)
+        self.assertEqual(' indented code\n', tokens[0].children[0].children[0].children[0].content)
+        self.assertIsInstance(tokens[0].children[0].children[1], block_token.Paragraph)
+        self.assertIsInstance(tokens[0].children[0].children[2], block_token.BlockCode)
+
 
 class TestList(unittest.TestCase):
     def test_different_markers(self):
@@ -335,27 +418,67 @@ class TestFootnote(unittest.TestCase):
         token = block_token.Document(lines)
         self.assertEqual(token.footnotes, {"key 1": ("value1", "title1"),
                                            "key 2": ("value2", "title2")})
-    
-    # this tests an edge case, it shouldn't occur in normal documents
-    def test_parse_with_para_right_after(self):
+
+    def test_parse_with_space_in_every_part(self):
+        lines = ['[Foo bar]:\n',
+                 '<my url>\n',
+                 '\'my title\'\n']
+        token = block_token.Document(lines)
+        self.assertEqual(set(token.footnotes.values()), set({ ("my url", "my title") }))
+
+    def test_parse_title_must_be_separated_from_link_destination(self):
+        lines = ['[foo]: <bar> (baz)\n']
+        token = block_token.Document(lines)
+        self.assertEqual(set(token.footnotes.values()), set({ ("bar", "baz") }))
+
+        lines = ['[foo]: <bar>(baz)\n']
+        token = block_token.Document(lines)
+        self.assertEqual(len(token.footnotes), 0)
+
+    # this tests an edge case, it shouldn't occur in normal documents:
+    # "[key 2]" is part of the paragraph above it, because a link reference definitions cannot interrupt a paragraph.
+    def test_footnote_followed_by_paragraph(self):
         lines = ['[key 1]: value1\n',
-                 # 'something1\n', # if uncommented,
-                 #     this and the next line should be treated as a paragraph
-                 #     - this line gets skipped instead now
+                 'something1\n',
                  '[key 2]: value2\n',
                  'something2\n',
                  '\n',
-                 '[key 3]: value3\r\n', # '\r', or any other whitespace
+                 '[key 3]: value3\r\n', # '\r', or any other whitespace may follow on the same line
                  'something3\n']
         token = block_token.Document(lines)
         self.assertEqual(token.footnotes, {"key 1": ("value1", ""),
-                                           "key 2": ("value2", ""),
                                            "key 3": ("value3", "")})
         self.assertEqual(len(token.children), 2)
         self.assertIsInstance(token.children[0], block_token.Paragraph)
-        self.assertEqual(token.children[0].children[0].content, "something2")
+        self.assertEqual(len(token.children[0].children), 5) # something1, <line break>, [key 2]: value2, <line break>, something2
+        self.assertEqual(token.children[0].children[2].content, "[key 2]: value2")
         self.assertEqual(token.children[1].children[0].content, "something3")
 
+    def test_content_after_title_not_allowed(self):
+        lines = ['[foo]: /url\n',
+                 '"title" ok\n']
+        token = block_token.Document(lines)
+        self.assertEqual(token.footnotes, {"foo": ("/url", "")})
+        self.assertEqual(len(token.children), 1)
+        self.assertIsInstance(token.children[0], block_token.Paragraph)
+        self.assertEqual(token.children[0].children[0].content, "\"title\" ok")
+
+    def test_footnotes_may_not_have_too_much_leading_space(self):
+        lines = ['   [link]: /bla\n',
+                 '    [i-am-block-actually]: /foo\n',
+                 'paragraph\n',
+                 '\n',
+                 '\t[i-am-block-too]: /foo\n']
+        token = block_token.Document(lines)
+        self.assertEqual(token.footnotes, {"link": ("/bla", "")})
+        self.assertEqual(len(token.children), 3)
+        self.assertIsInstance(token.children[0], block_token.BlockCode)
+        self.assertEqual(token.children[0].children[0].content, "[i-am-block-actually]: /foo\n")
+        self.assertIsInstance(token.children[1], block_token.Paragraph)
+        self.assertEqual(token.children[1].children[0].content, "paragraph")
+        self.assertIsInstance(token.children[2], block_token.BlockCode)
+        self.assertEqual(token.children[2].children[0].content, "[i-am-block-too]: /foo\n")
+
     def test_parse_opening_bracket_as_paragraph(self): # ... and no error is raised
         lines = ['[\n']
         token = block_token.Document(lines)
@@ -411,3 +534,73 @@ class TestContains(unittest.TestCase):
         self.assertTrue('heading' in token)
         self.assertTrue('code' in token)
         self.assertFalse('foo' in token)
+
+
+class TestHTMLBlock(unittest.TestCase):
+    def setUp(self):
+        block_token.add_token(block_token.HTMLBlock)
+        self.addCleanup(block_token.reset_tokens)
+
+    def test_textarea_block_may_contain_blank_lines(self):
+        lines = ['<textarea>\n',
+                 '\n',
+                 '*foo*\n',
+                 '\n',
+                 '_bar_\n',
+                 '\n',
+                 '</textarea>\n']
+        document = block_token.Document(lines)
+        tokens = document.children
+        self.assertEqual(1, len(tokens))
+        self.assertIsInstance(tokens[0], block_token.HTMLBlock)
+
+
+class TestLeafBlockTokenContentProperty(unittest.TestCase):
+    def setUp(self):
+        block_token.add_token(block_token.HTMLBlock)
+        self.addCleanup(block_token.reset_tokens)
+
+    def test_code_fence(self):
+        lines = ['```\n',
+                 'line 1\n',
+                 'line 2\n',
+                 '```\n']
+        document = block_token.Document(lines)
+        tokens = document.children
+        self.assertEqual(1, len(tokens))
+        self.assertIsInstance(tokens[0], block_token.CodeFence)
+
+        # option 1: direct access to the content
+        self.assertEqual('line 1\nline 2\n', tokens[0].children[0].content)
+
+        # option 2: using property getter to access the content
+        self.assertEqual('line 1\nline 2\n', tokens[0].content)
+
+    def test_block_code(self):
+        lines = ['    line 1\n',
+                 '    line 2\n']
+        document = block_token.Document(lines)
+        tokens = document.children
+        self.assertEqual(1, len(tokens))
+        self.assertIsInstance(tokens[0], block_token.BlockCode)
+
+        # option 1: direct access to the content
+        self.assertEqual('line 1\nline 2\n', tokens[0].children[0].content)
+
+        # option 2: using property getter to access the content
+        self.assertEqual('line 1\nline 2\n', tokens[0].content)
+
+    def test_html_block(self):
+        lines = ['<div>\n',
+                 'text\n'
+                 '</div>\n']
+        document = block_token.Document(lines)
+        tokens = document.children
+        self.assertEqual(1, len(tokens))
+        self.assertIsInstance(tokens[0], block_token.HTMLBlock)
+
+        # option 1: direct access to the content
+        self.assertEqual(''.join(lines).strip(), tokens[0].children[0].content)
+
+        # option 2: using property getter to access the content
+        self.assertEqual(''.join(lines).strip(), tokens[0].content)
