@@ -1,7 +1,7 @@
 from unittest import TestCase
 from mistletoe import block_token
 from mistletoe.block_token import Document, Heading
-from contrib.toc_renderer import TOCRenderer
+from mistletoe.contrib.toc_renderer import TOCRenderer
 
 class TestTOCRenderer(TestCase):
     def test_parse_rendered_heading(self):
@@ -13,7 +13,7 @@ class TestTOCRenderer(TestCase):
         renderer = TOCRenderer()
         Heading.start('### some *text*\n')
         token = Heading(Heading.read(iter(['foo'])))
-        rendered_heading = renderer.render_heading(token)
+        renderer.render_heading(token)
         self.assertEqual(renderer._headings[0], (3, 'some text'))
 
     def test_depth(self):
