@@ -1,6 +1,7 @@
 from unittest import TestCase, mock
-import mistletoe.latex_token as latex_token
+import mistletoe.latex_renderer
 from mistletoe.latex_renderer import LaTeXRenderer
+from mistletoe import markdown
 
 
 class TestLaTeXRenderer(TestCase):
@@ -28,8 +29,19 @@ class TestLaTeXRenderer(TestCase):
 
     def test_inline_code(self):
         func_path = 'mistletoe.latex_renderer.LaTeXRenderer.render_raw_text'
-        with mock.patch(func_path, return_value='inner'):
-            self._test_token('InlineCode', '\\verb|inner|')
+
+        for content, output in {'inner': '\\verb|inner|',
+                                'a + b': '\\verb|a + b|',
+                                'a | b': '\\verb!a | b!',
+                                '|ab!|': '\\verb"|ab!|"',
+                               }.items():
+            with mock.patch(func_path, return_value=content):
+                self._test_token('InlineCode', output, content=content)
+
+        content = mistletoe.latex_renderer.verb_delimiters
+        with self.assertRaises(RuntimeError):
+            with mock.patch(func_path, return_value=content):
+                self._test_token('InlineCode', None, content=content)
 
     def test_strikethrough(self):
         self._test_token('Strikethrough', '\\sout{inner}')
@@ -112,6 +124,14 @@ class TestLaTeXRenderer(TestCase):
         self._test_token('Document', output, footnotes={})
 
 
+class TestHtmlEntity(TestCase):
+    def test_html_entity(self):
+        self.assertIn('hello \\& goodbye', markdown('hello &amp; goodbye', LaTeXRenderer))
+
+    def test_html_entity_in_link_target(self):
+        self.assertIn('\\href{foo}{hello}', markdown('[hello](f&#111;&#111;)', LaTeXRenderer))
+
+
 class TestLaTeXFootnotes(TestCase):
     def setUp(self):
         self.renderer = LaTeXRenderer()
